/* Tactical Analysis for AIs
   Copyright (C) 2005 Eric A. McDonald.

Xconq is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.  See the file COPYING.  */

/*! \file 
    \brief Tactical Analysis for AIs

    Part of the AI API, Level 3. 

    Provides a tactical level AI implementation.

    \note Nothing in this file should ever be required to implement an AI; 
	    everything here is optional.

*/

namespace Xconq {
namespace AI {

//! Enumeration of tactical concerns.
enum TacticalConcern {
    //! No tactical concern.
    TC_NONE = 0,
    //! Immediate danger.
    TC_DANGER_IMMEDIATE,
    //! Total number of tactical concern types.
    TC_Total
};

/** Task Inspection **/

//! How many friendly units plan to construct at given location?
extern int n_plan_to_construct_at(
    Side *side, int x, int y, short *p_utypes = NULL);

/** Resource Contention Assessement **/

//! Would any units be in contention with an utype at a given location?
extern int resource_contention_with_any(int u, int x, int y, Side *side);

/** Threat Assessment and Handling **/

//! Handle immediate danger to given unit.
extern TacticalConcern handle_immediate_danger(Unit *unit);

/** Overall Tactical Assessment and Handling **/

//! Handle tactical situation of given unit.
extern TacticalConcern handle_tactical_situation(Unit *unit);

} // namespace Xconq::AI
} // namespace Xconq
